from enum import Enum
from pathlib import Path
from typing import Optional

import typer

from reprexlite.config import ParsingMethod, ReprexConfig
from reprexlite.exceptions import InputSyntaxError, IPythonNotFoundError
from reprexlite.formatting import formatter_registry
from reprexlite.reprexes import Reprex
from reprexlite.version import __version__

app = typer.Typer()


def version_callback(version: bool):
    """Print reprexlite version to console."""
    if version:
        print(__version__)
        raise typer.Exit()


Venue = Enum(  # type: ignore
    "Venue", names={v.upper(): v for v in formatter_registry.keys()}, type=str
)
Venue.__doc__ = """Enum for valid venue options."""


def get_help(key: str):
    return ReprexConfig.get_help(key).replace("_", "-")


@app.command()
def main(
    editor: Optional[str] = typer.Option(
        None,
        "--editor",
        "-e",
        help=(
            "Specify editor to open in. Can be full path to executable or name to be searched on "
            "system search path. If None, will use Click's automatic editor detection. This will "
            "typically use the editor set to environment variable VISUAL or EDITOR. If value is "
            "'ipython' and IPython is installed, this will launch the interactive IPython editor "
            "where all cells are automatically run through reprexlite."
        ),
    ),
    infile: Optional[Path] = typer.Option(
        None,
        "--infile",
        "-i",
        help="Read code from an input file instead of entering in an editor.",
    ),
    outfile: Optional[Path] = typer.Option(
        None, "--outfile", "-o", help="Write output to file instead of printing to console."
    ),
    # Formatting
    venue: Venue = typer.Option(
        "gh",
        "--venue",
        "-v",
        help=get_help("venue"),
    ),
    advertise: Optional[bool] = typer.Option(
        None,
        "--advertise/--no-advertise",
        help=get_help("advertise"),
        is_flag=False,
        show_default=False,
    ),
    session_info: bool = typer.Option(False, help=get_help("session_info")),
    style: bool = typer.Option(False, help=get_help("style")),
    prompt: str = typer.Option("", help=get_help("prompt")),
    continuation: str = typer.Option("", help=get_help("continuation")),
    comment: str = typer.Option("#>", help=get_help("comment")),
    keep_old_results: bool = typer.Option(False, help=get_help("keep_old_results")),
    # Parsing
    parsing_method: ParsingMethod = typer.Option("auto", help=get_help("parsing_method")),
    input_prompt: Optional[str] = typer.Option(None, help=get_help("input_prompt")),
    input_continuation: Optional[str] = typer.Option(None, help=get_help("input_continuation")),
    input_comment: Optional[str] = typer.Option(None, help=get_help("input_comment")),
    verbose: Optional[bool] = typer.Option(None, "--verbose"),
    # Callbacks
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show reprexlite version and exit.",
    ),
):
    """Render reproducible examples of Python code for sharing. Your code will be executed and, in
    the default output style, the results will be embedded as comments below their associated
    lines.

    By default, your system's default command-line text editor will open for you to type or paste
    in your code. This editor can be changed by setting the VISUAL or EDITOR environment variable,
    or by explicitly passing in the --editor program. You can instead specify an input file with
    the --infile / -i option. If IPython is installed, an interactive IPython editor can also be
    launched using the --ipython flag.

    Additional markup will be added that is appropriate to the choice of venue formatting. For
    example, for the default `gh` venue for GitHub Flavored Markdown, the final reprex output will
    look like:

    \b
    ----------------------------------------
    ```python
    arr = [1, 2, 3, 4, 5]
    [x + 1 for x in arr]
    #> [2, 3, 4, 5, 6]
    max(arr) - min(arr)
    #> 4
    ```
    \b
    <sup>Created at 2021-02-27 00:13:55 PST by [reprexlite](https://github.com/jayqi/reprexlite) v{version}</sup>
    ----------------------------------------

    \b
    The supported venue formats are:
    \b
    - gh : GitHub Flavored Markdown
    - so : StackOverflow, alias for gh
    - ds : Discourse, alias for gh
    - html : HTML
    - py : Python script
    - rtf : Rich Text Format
    - slack : Slack
    """  # noqa: E501

    config = ReprexConfig(
        venue=venue.value,
        advertise=advertise,
        session_info=session_info or False,
        style=style or False,
        prompt=prompt,
        continuation=continuation,
        comment=comment,
        parsing_method=parsing_method.value,
        input_prompt=input_prompt,
        input_continuation=input_continuation,
        input_comment=input_comment,
        keep_old_results=keep_old_results or False,
    )

    if editor and editor.lower() == "ipython":
        try:
            from reprexlite.ipython import ReprexTerminalIPythonApp
        except IPythonNotFoundError:
            print(
                "IPythonNotFoundError: ipython is required to be installed to use the IPython "
                "interactive editor."
            )
            raise typer.Exit(code=1)
        ReprexTerminalIPythonApp.set_reprex_config(config)
        ReprexTerminalIPythonApp.launch_instance(argv=[])
        raise typer.Exit()
    elif infile:
        with infile.open("r") as fp:
            input = fp.read()
    else:
        input = typer.edit(editor=editor) or ""

    if verbose:
        typer.echo(config)

    try:
        r = Reprex.from_input(input=input, config=config)
    except InputSyntaxError as e:
        print("ERROR: reprexlite has encountered an error while evaluating your input.")
        print(e)
        raise typer.Exit(1) from e

    if outfile:
        with outfile.open("w") as fp:
            fp.write(r.format(terminal=False))
        print(f"Wrote rendered reprex to {outfile}")
    else:
        print(r.format(terminal=True), end="")

    return r


if main.__doc__:
    main.__doc__ = main.__doc__.format(version=__version__)



from contextlib import redirect_stdout
import dataclasses
from io import StringIO
from itertools import chain
import os
from pathlib import Path
from pprint import pformat
import traceback
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

try:
    from typing import Self  # type: ignore  # Python 3.11+
except ImportError:
    from typing_extensions import Self

import libcst as cst

from reprexlite.config import ParsingMethod, ReprexConfig
from reprexlite.exceptions import BlackNotFoundError, InputSyntaxError, UnexpectedError
from reprexlite.formatting import formatter_registry
from reprexlite.parsing import LineType, auto_parse, parse


@dataclasses.dataclass
class RawResult:
    """Class that holds the result of evaluated code. Use `str(...)` on an instance to produce a
    pretty-formatted comment block representation of the result.

    Attributes:
        config (ReprexConfig): Configuration for formatting and parsing
        raw (Any): Some Python object that is the raw return value of evaluated Python code.
        stdout (str): Standard output from evaluated Python code.
    """

    config: ReprexConfig
    raw: Any
    stdout: Optional[str]

    def __str__(self) -> str:
        if not self:
            raise UnexpectedError("Should not print a RawResult if it tests False.")
        lines = []
        if self.stdout:
            lines.extend(self.stdout.split("\n"))
        if self.raw is not None:
            lines.extend(pformat(self.raw, indent=2, width=77).split("\n"))
        if self.config.comment:
            return "\n".join(self.config.comment + " " + line for line in lines)
        else:
            return "\n".join(lines)

    def __bool__(self) -> bool:
        """Tests whether instance contains anything to print."""
        return not (self.raw is None and self.stdout is None)

    def __repr__(self) -> str:
        return (
            f"<RawResult '{to_snippet(str(self.raw), 10)}' '{to_snippet(str(self.stdout), 10)}'>"
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, RawResult):
            return self.raw == other.raw and self.stdout == other.stdout
        else:
            return NotImplemented


@dataclasses.dataclass
class ParsedResult:
    """Class that holds parsed result from reading a reprex.

    Attributes:
        config (ReprexConfig): Configuration for formatting and parsing
        lines (List[str]): String content of result parsed from a reprex
    """

    config: ReprexConfig
    lines: List[str]

    def __str__(self) -> str:
        if not self:
            raise UnexpectedError("Should not print a ParsedResult if it tests False.")
        return "\n".join(self.prefix * 2 + line for line in self.lines)

    def as_result_str(self) -> str:
        return "\n".join(self.prefix + line for line in self.lines)

    @property
    def prefix(self) -> str:
        if self.config.comment:
            return self.config.comment + " "
        else:
            return ""

    def __bool__(self) -> bool:
        """Tests whether instance contains anything to print."""
        return bool(self.lines)

    def __repr__(self) -> str:
        joined = "\\n".join(self.lines)
        return f"<ParsedResult '{to_snippet(joined, 10)}'>"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ParsedResult):
            return self.lines == other.lines
        elif isinstance(other, RawResult):
            if not bool(self) and not bool(other):
                return True
            return bool(self) and bool(other) and self.as_result_str() == str(other)
        else:
            return NotImplemented


@dataclasses.dataclass
class Statement:
    """Dataclass that holds a LibCST parsed statement. of code.

    Attributes:
        config (ReprexConfig): Configuration for formatting and parsing
        stmt (Union[libcst.SimpleStatementLine, libcst.BaseCompoundStatement]): LibCST parsed
            statement.
    """

    config: ReprexConfig
    stmt: Union[cst.SimpleStatementLine, cst.BaseCompoundStatement, cst.EmptyLine]

    def evaluate(self, scope: dict) -> RawResult:
        """Evaluate code statement and produce a RawResult dataclass instance.

        Args:
            scope (dict): scope to use for evaluation

        Returns:
            RawResult: Dataclass instance holding evaluation results.
        """
        if isinstance(self.stmt, cst.EmptyLine):
            return RawResult(config=self.config, raw=None, stdout=None)

        if "__name__" not in scope:
            scope["__name__"] = "__reprex__"
        stdout_io = StringIO()
        try:
            with redirect_stdout(stdout_io):
                try:
                    # Treat as a single expression
                    result = eval(self.code.strip(), scope)
                except SyntaxError:
                    # Treat as a statement
                    exec(self.code.strip(), scope)
                    result = None
            stdout = stdout_io.getvalue().strip()
        except Exception as exc:
            result = None
            # Skip first step of traceback, since that is this evaluate method
            if exc.__traceback__ is not None:
                tb = exc.__traceback__.tb_next
                stdout = (
                    "Traceback (most recent call last):\n"
                    + "".join(line for line in traceback.format_tb(tb))
                    + f"{type(exc).__name__}: {exc}"
                )
        finally:
            stdout_io.close()

        return RawResult(config=self.config, raw=result, stdout=stdout or None)

    @property
    def raw_code(self) -> str:
        """Raw code of contained statement as a string."""
        if isinstance(self.stmt, cst.EmptyLine):
            return cst.Module(body=[], header=[self.stmt]).code.rstrip()
        return cst.Module(body=[self.stmt]).code.rstrip()

    @property
    def code(self) -> str:
        """Code of contained statement. May be autoformatted depending on configuration."""
        code = self.raw_code
        if self.config.style:
            try:
                from black import Mode, format_str
            except ModuleNotFoundError as e:
                if e.name == "black":
                    raise BlackNotFoundError("Must install black to restyle code.", name="black")
                else:
                    raise

            code = format_str(code, mode=Mode())
        return code

    def __str__(self) -> str:
        out = self.code
        if self.config.prompt:
            # Add prompt and continuation prefixes to lines
            lines = out.split("\n")
            primary_found = False
            out = ""
            for line in lines:
                if line.strip() == "":
                    # Whitespace line
                    out += f"{self.config.prompt} " + line + "\n"
                elif line.startswith("#"):
                    # Comment line
                    out += f"{self.config.prompt} " + line + "\n"
                else:
                    # Code line
                    if not primary_found:
                        out += f"{self.config.prompt} " + line + "\n"
                        primary_found = True
                    else:
                        out += f"{self.config.continuation} " + line + "\n"
        return out.rstrip()

    def __bool__(self) -> bool:
        """Tests whether this instance contains anything to print. Always true for Statement."""
        return True

    def __repr__(self) -> str:
        return f"<Statement '{to_snippet(self.code, 10)}'>"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Statement):
            return self.raw_code == other.raw_code
        return NotImplemented


@dataclasses.dataclass
class Reprex:
    """Dataclass for a reprex, which holds Python code and results from evaluation.

    Attributes:
        config (ReprexConfig): Configuration for formatting and parsing
        statements (List[Statement]): List of parsed Python code statements
        results (List[RawResult]): List of results evaluated from statements
        old_results (List[ParsedResult]): List of any old results parsed from input code
        scope (Dict[str, Any]): Dictionary holding the scope that the reprex was evaluated in
    """

    config: ReprexConfig
    statements: List[Statement]
    results: List[RawResult]
    old_results: List[ParsedResult]
    scope: Dict[str, Any]

    def __post_init__(self) -> None:
        if not (len(self.statements) == len(self.results) == len(self.old_results)):
            raise UnexpectedError(
                "statements, results, and old_results should all be the same length. "
                f"Got: {(len(self.statements), len(self.results), len(self.old_results))}."
            )

    @classmethod
    def from_input(
        cls,
        input: str,
        config: Optional[ReprexConfig] = None,
        scope: Optional[Dict[str, Any]] = None,
    ) -> Self:
        """Create a Reprex instance from parsing and evaluating code from a string.

        Args:
            input (str): Input code
            config (Optional[ReprexConfig], optional): Configuration. Defaults to None, which will
                use default settings.
            scope (Optional[Dict[str, Any]], optional): Dictionary holding scope that the parsed
                code will be evaluated with. Defaults to None, which will create an empty
                dictionary.

        Returns:
            Reprex: New instance of Reprex.
        """
        if config is None:
            config = ReprexConfig()
        if config.parsing_method == ParsingMethod.AUTO:
            lines = list(auto_parse(input))
        elif config.parsing_method == ParsingMethod.DECLARED:
            lines = list(
                parse(
                    input,
                    prompt=config.resolved_input_prompt,
                    continuation=config.resolved_input_continuation,
                    comment=config.resolved_input_comment,
                )
            )
        else:
            raise UnexpectedError(  # pragma: nocover
                f"Parsing method {config.parsing_method} is not implemented."
            )
        return cls.from_input_lines(lines, config=config, scope=scope)

    @classmethod
    def from_input_lines(
        cls,
        lines: Sequence[Tuple[str, LineType]],
        config: Optional[ReprexConfig] = None,
        scope: Optional[Dict[str, Any]] = None,
    ) -> Self:
        """Creates a Reprex instance from the output of [parse][reprexlite.parsing.parse].

        Args:
            lines (Sequence[Tuple[str, LineType]]): Output from parse.
            config (Optional[ReprexConfig], optional): Configuration. Defaults to None, which will
                use default settings.
            scope (Optional[Dict[str, Any]], optional): Dictionary holding scope that the parsed
                code will be evaluated with. Defaults to None, which will create an empty
                dictionary.

        Returns:
            Reprex: New instance of Reprex.
        """
        if config is None:
            config = ReprexConfig()
        statements: List[Statement] = []
        old_results: List[ParsedResult] = []
        current_code_block: List[str] = []
        current_result_block: List[str] = []
        try:
            for line_content, line_type in lines:
                if line_type is LineType.CODE:
                    # Flush results
                    if current_result_block:
                        old_results += [ParsedResult(config=config, lines=current_result_block)]
                        current_result_block = []
                    # Append line to current code
                    current_code_block.append(line_content)
                elif line_type is LineType.RESULT:
                    # Flush code
                    if current_code_block:
                        # Parse code and create Statements
                        tree: cst.Module = cst.parse_module("\n".join(current_code_block))
                        new_statements = (
                            [Statement(config=config, stmt=stmt) for stmt in tree.header]
                            + [Statement(config=config, stmt=stmt) for stmt in tree.body]
                            + [Statement(config=config, stmt=stmt) for stmt in tree.footer]
                        )
                        statements += new_statements
                        # Pad results with empty results, 1 fewer because of current_result_block
                        old_results += [ParsedResult(config=config, lines=[])] * (
                            len(new_statements) - 1
                        )
                        # Reset current code block
                        current_code_block = []
                    # Append line to current results
                    current_result_block.append(line_content)
            # Flush code
            if current_code_block:
                if all(not line for line in current_code_block):
                    # Case where all lines are whitespace: strip and don't add
                    new_statements = []
                else:
                    # Parse code and create Statements
                    tree: cst.Module = cst.parse_module(  # type: ignore[no-redef]
                        "\n".join(current_code_block)
                    )
                    new_statements = (
                        [Statement(config=config, stmt=stmt) for stmt in tree.header]
                        + [Statement(config=config, stmt=stmt) for stmt in tree.body]
                        + [Statement(config=config, stmt=stmt) for stmt in tree.footer]
                    )
                # Pad results with empty results, 1 fewer because of current_result_block
                statements += new_statements
                old_results += [ParsedResult(config=config, lines=[])] * (len(new_statements) - 1)
            # Flush results
            if current_result_block:
                old_results += [ParsedResult(config=config, lines=current_result_block)]
            # Pad results to equal length
            old_results += [ParsedResult(config=config, lines=[])] * (
                len(statements) - len(old_results)
            )

            # Evaluate for new results
            if scope is None:
                scope = {}
            results = [statement.evaluate(scope=scope) for statement in statements]
            return cls(
                config=config,
                statements=statements,
                results=results,
                old_results=old_results,
                scope=scope,
            )
        except cst.ParserSyntaxError as e:
            raise InputSyntaxError(str(e)) from e

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Reprex):
            return (
                self.config == other.config
                and all(left == right for left, right in zip(self.statements, other.statements))
                and all(left == right for left, right in zip(self.results, other.results))
                and all(left == right for left, right in zip(self.old_results, other.old_results))
            )
        else:
            return NotImplemented

    def __str__(self) -> str:
        if self.config.keep_old_results:
            lines = chain.from_iterable(zip(self.statements, self.old_results, self.results))
        else:
            lines = chain.from_iterable(zip(self.statements, self.results))
        out = "\n".join(str(line) for line in lines if line)
        if not out.endswith("\n"):
            out += "\n"
        return out

    @property
    def results_match(self) -> bool:
        """Whether results of evaluating code match old results parsed from input."""
        return all(
            result == old_result for result, old_result in zip(self.results, self.old_results)
        )

    def format(self, terminal: bool = False) -> str:
        out = str(self)
        if terminal:
            try:
                from pygments import highlight
                from pygments.formatters import Terminal256Formatter
                from pygments.lexers import PythonLexer

                out = highlight(out, PythonLexer(), Terminal256Formatter(style="friendly"))
            except ModuleNotFoundError:
                pass
        formatter = formatter_registry[self.config.venue]
        return formatter.format(
            out.strip(), advertise=self.config.advertise, session_info=self.config.session_info
        )

    def __repr__(self) -> str:
        return f"<Reprex ({len(self.statements)}) '{to_snippet(str(self), 10)}'>"

    def _repr_html_(self) -> str:
        """HTML representation. Used for rendering in Jupyter."""
        out = []
        try:
            from pygments import highlight
            from pygments.formatters import HtmlFormatter
            from pygments.lexers import PythonLexer

            formatter = HtmlFormatter(style="friendly", wrapcode=True)
            out.append(f"<style>{formatter.get_style_defs('.highlight')}</style>")
            out.append(highlight(self.format(), PythonLexer(), formatter))
        except ModuleNotFoundError:
            out.append(f"<pre><code>{self.format()}</code></pre>")
        return "\n".join(out)


def to_snippet(s: str, n: int) -> str:
    if len(s) <= n:
        return rf"{s}"
    else:
        return rf"{s[:n]}..."


def reprex(
    input: str,
    outfile: Optional[Union[str, os.PathLike]] = None,
    print_: bool = True,
    terminal: bool = False,
    config: Optional[ReprexConfig] = None,
    **kwargs,
) -> Reprex:
    """A convenient functional interface to render reproducible examples of Python code for
    sharing. This function will evaluate your code and, by default, print out your code with the
    evaluated results embedded as comments, formatted with additional markup appropriate to the
    sharing venue set by the `venue` keyword argument. The function returns an instance of
    [`Reprex`][reprexlite.reprexes.Reprex] which holds the relevant data.

    For example, for the `gh` venue for GitHub Flavored Markdown, you'll get a reprex whose
    formatted output looks like:

    ````
    ```python
    x = 2
    x + 2
    #> 4
    ```

    <sup>Created at 2021-02-15 16:58:47 PST by [reprexlite](https://github.com/jayqi/reprexlite)</sup>
    ````


    Args:
        input (str): Input code to create a reprex for.
        outfile (Optional[str | os.PathLike]): If provided, path to write formatted reprex
            output to. Defaults to None, which does not write to any file.
        print_ (bool): Whether to print formatted reprex output to console.
        terminal (bool): Whether currently in a terminal. If true, will automatically apply code
            highlighting if pygments is installed.
        config (Optional[ReprexConfig]): Instance of the configuration dataclass. Default of none
            will instantiate one with default values.
        **kwargs: Configuration options from [ReprexConfig][reprexlite.config.ReprexConfig]. Any
            provided values will override values from provided config or the defaults.

    Returns:
        (Reprex) Reprex instance
    """  # noqa: E501

    if config is None:
        config = ReprexConfig(**kwargs)
    else:
        config = dataclasses.replace(config, **kwargs)

    config = ReprexConfig(**kwargs)
    if config.venue in ["html", "rtf"]:
        # Don't screw up output file or lexing for HTML and RTF with terminal syntax highlighting
        terminal = False
    r = Reprex.from_input(input, config=config)
    output = r.format(terminal=terminal)
    if outfile is not None:
        with Path(outfile).open("w") as fp:
            fp.write(r.format(terminal=False))
    if print_:
        print(output)
    return r



from abc import ABC, abstractmethod
import dataclasses
from datetime import datetime
from textwrap import dedent
from typing import ClassVar, Dict, Optional, Type

from reprexlite.exceptions import NotAFormatterError, PygmentsNotFoundError
from reprexlite.session_info import SessionInfo
from reprexlite.version import __version__


@dataclasses.dataclass
class FormatterMetadata:
    example: Optional[str]
    venues: Dict[str, str] = dataclasses.field(default_factory=lambda: dict())


class Formatter(ABC):
    """Abstract base class for a reprex formatter. Concrete subclasses should implement the
    formatting logic appropriate to a specific venue for sharing. Call `str(...)` on an instance
    to return the formatted reprex.

    Attributes:
        default_advertise (bool): Whether to render reprexlite advertisement by default
        meta (FormatterMeta): Contains metadata for the formatter, such as label text and an
            example
    """

    default_advertise: ClassVar[bool] = True
    meta: ClassVar[FormatterMetadata]

    @classmethod
    @abstractmethod
    def format(
        cls, reprex_str: str, advertise: Optional[bool] = None, session_info: bool = False
    ) -> str:
        """Format a reprex string for a specific sharing venue.

        Args:
            reprex_str (str): String containing rendered reprex output.
            advertise (Optional[bool], optional): Whether to include the advertisement for
                reprexlite. Defaults to None, which uses a per-formatter default.
            session_info (bool, optional): Whether to include detailed session information.
                Defaults to False.

        Returns:
            str: String containing formatted reprex code. Ends with newline.
        """


formatter_registry: Dict[str, Type[Formatter]] = {}
"""Registry of formatters keyed by venue keywords."""


def register_formatter(venue: str, label: str):
    """Decorator that registers a formatter implementation.

    Args:
        venue (str): Venue keyword that formatter will be registered to.
        label (str): Short human-readable label explaining the venue.
    """

    def registrar(cls):
        global formatter_registry
        if not isinstance(cls, type) or not issubclass(cls, Formatter):
            raise NotAFormatterError("Only subclasses of Formatter can be registered.")
        formatter_registry[venue] = cls
        cls.meta.venues[venue] = label
        return cls

    return registrar


@register_formatter(venue="ds", label="Discourse (alias for 'gh')")
@register_formatter(venue="so", label="StackOverflow (alias for 'gh')")
@register_formatter(venue="gh", label="Github Flavored Markdown")
class GitHubFormatter(Formatter):
    """Formatter for rendering reprexes in GitHub Flavored Markdown."""

    default_advertise = True
    meta = FormatterMetadata(
        example=dedent(
            """\
            ```python
            2+2
            #> 4
            ```
            """
        )
    )

    @classmethod
    def format(
        cls, reprex_str: str, advertise: Optional[bool] = None, session_info: bool = False
    ) -> str:
        if advertise is None:
            advertise = cls.default_advertise
        out = []
        out.append("```python")
        out.append(reprex_str)
        out.append("```")
        if advertise:
            out.append("\n" + Advertisement().markdown())
        if session_info:
            out.append("\n<details><summary>Session Info</summary>")
            out.append("```text")
            out.append(str(SessionInfo()))
            out.append("```")
            out.append("</details>")
        return "\n".join(out) + "\n"


@register_formatter(venue="html", label="HTML")
class HtmlFormatter(Formatter):
    """Formatter for rendering reprexes in HTML. If optional dependency Pygments is
    available, the rendered HTML will have syntax highlighting for the Python code."""

    default_advertise = True
    meta = FormatterMetadata(
        example=dedent(
            """\
            <pre><code>2+2
            #> 4</code></pre>
            """
        )
    )

    @classmethod
    def format(
        cls, reprex_str: str, advertise: Optional[bool] = None, session_info: bool = False
    ) -> str:
        if advertise is None:
            advertise = cls.default_advertise
        out = []
        try:
            from pygments import highlight
            from pygments.formatters import HtmlFormatter
            from pygments.lexers import PythonLexer

            formatter = HtmlFormatter(
                style="friendly", lineanchors=True, linenos=True, wrapcode=True
            )
            out.append(f"<style>{formatter.get_style_defs('.highlight')}</style>")
            out.append(highlight(str(reprex_str), PythonLexer(), formatter))
        except ImportError:
            out.append(f"<pre><code>{reprex_str}</code></pre>")

        if advertise:
            out.append(Advertisement().html().strip())
        if session_info:
            out.append("<details><summary>Session Info</summary>")
            out.append(f"<pre><code>{SessionInfo()}</code></pre>")
            out.append("</details>")
        return "\n".join(out) + "\n"


@register_formatter(venue="py", label="Python script")
class PyScriptFormatter(Formatter):
    """Formatter for rendering reprexes as a Python script."""

    default_advertise = False
    meta = FormatterMetadata(
        example=dedent(
            """\
            2+2
            #> 4
            """
        )
    )

    @classmethod
    def format(
        cls, reprex_str: str, advertise: Optional[bool] = None, session_info: bool = False
    ) -> str:
        if advertise is None:
            advertise = cls.default_advertise
        out = [str(reprex_str)]
        if advertise:
            out.append("\n" + Advertisement().code_comment())
        if session_info:
            out.append("")
            sess_lines = str(SessionInfo()).split("\n")
            out.extend("# " + line for line in sess_lines)
        return "\n".join(out) + "\n"


@register_formatter(venue="rtf", label="Rich Text Format")
class RtfFormatter(Formatter):
    """Formatter for rendering reprexes in Rich Text Format."""

    default_advertise = False
    meta = FormatterMetadata(example=None)

    @classmethod
    def format(
        cls, reprex_str: str, advertise: Optional[bool] = None, session_info: bool = False
    ) -> str:
        if advertise is None:
            advertise = cls.default_advertise
        try:
            from pygments import highlight
            from pygments.formatters import RtfFormatter
            from pygments.lexers import PythonLexer
        except ModuleNotFoundError as e:
            if e.name == "pygments":
                raise PygmentsNotFoundError(
                    "Pygments is required for RTF output.", name="pygments"
                )
            else:
                raise

        out = str(reprex_str)
        if advertise:
            out += "\n\n" + Advertisement().text()
        if session_info:
            out += "\n\n" + str(SessionInfo())
        return highlight(out, PythonLexer(), RtfFormatter()) + "\n"


@register_formatter(venue="slack", label="Slack")
class SlackFormatter(Formatter):
    """Formatter for rendering reprexes as Slack markup."""

    default_advertise = False
    meta = FormatterMetadata(
        example=dedent(
            """\
            ```
            2+2
            #> 4
            ```
            """
        )
    )

    @classmethod
    def format(
        cls, reprex_str: str, advertise: Optional[bool] = None, session_info: bool = False
    ) -> str:
        if advertise is None:
            advertise = cls.default_advertise
        out = []
        out.append("```")
        out.append(str(reprex_str))
        out.append("```")
        if advertise:
            out.append("\n" + Advertisement().text())
        if session_info:
            out.append("\n```")
            out.append(str(SessionInfo()))
            out.append("```")
        return "\n".join(out) + "\n"


class Advertisement:
    """Class for generating the advertisement note for reprexlite.

    Attributes:
        timestamp (str): Timestamp of instance instantiation
        version (str): Version of reprexlite
    """

    pkg = "reprexlite"
    url = "https://github.com/jayqi/reprexlite"

    def __init__(self):
        self.timestamp = datetime.now().astimezone().strftime("%Y-%m-%d %H:%M %Z")
        self.version = f"v{__version__}"

    def markdown(self) -> str:
        """Render reprexlite advertisement in GitHub Flavored Markdown."""
        return f"<sup>Created at {self.timestamp} by [{self.pkg}]({self.url}) {self.version}</sup>"

    def html(self) -> str:
        """Render reprexlite advertisement in HTML."""
        return (
            f"<p><sup>Created at {self.timestamp} by "
            f'<a href="{self.url}">{self.pkg}</a> {self.version}</sup></p>'
        )

    def code_comment(self) -> str:
        """Render reprexlite advertisement as a comment in Python code."""
        return f"# {self.text()}"

    def text(self) -> str:
        """Render reprexlite advertisement in plain text."""
        return f"Created at {self.timestamp} by {self.pkg} {self.version} <{self.url}>"



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from reprexlite.exceptions import (
    InvalidParsingMethodError,
    InvalidVenueError,
    PromptLengthMismatchError,
)
from reprexlite.formatting import formatter_registry


class ParsingMethod(str, Enum):
    """Methods for parsing input strings.

    Args:
        AUTO (str): Automatically identify reprex-style or doctest-style input.
        DECLARED (str): Use configured values for parsing.
    """

    AUTO = "auto"
    DECLARED = "declared"


@dataclass
class ReprexConfig:
    """Configuration dataclass for reprexlite. Used to configure input parsing and output
    formatting.
    """

    # Formatting
    venue: str = field(
        default="gh",
        metadata={
            "help": (
                "Key to identify the output venue that the reprex will be shared in. Used to "
                'select an appropriate formatter. See "Venues Formatting" documentation for '
                "formats included with reprexlite."
            )
        },
    )
    advertise: Optional[bool] = field(
        default=None,
        metadata={
            "help": (
                "Whether to include a footer that credits reprexlite. If unspecified, will depend "
                "on specified venue formatter's default."
            )
        },
    )
    session_info: bool = field(
        default=False,
        metadata={
            "help": (
                "Include details about session and environment that the reprex was generated with."
            )
        },
    )
    style: bool = field(
        default=False,
        metadata={
            "help": "Whether to autoformat code with black. Requires black to be installed."
        },
    )
    prompt: str = field(
        default="",
        metadata={"help": "Prefix to use as primary prompt for code lines."},
    )
    continuation: str = field(
        default="",
        metadata={"help": "Prefix to use as secondary prompt for continued code lines."},
    )
    comment: str = field(
        default="#>",
        metadata={"help": "Prefix to use for results returned by expressions."},
    )
    keep_old_results: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether to additionally include results of expressions detected in the original "
                "input when formatting the reprex output."
            )
        },
    )
    # Parsing
    parsing_method: str = field(
        default="auto",
        metadata={
            "help": (
                "Method for parsing input. 'auto' will automatically detect either default "
                "reprex-style input or standard doctest-style input. 'declared' will allow you to "
                "specify custom line prefixes. Values for 'prompt', 'continuation', and 'comment' "
                "will be used for both output formatting and input parsing, unless the associated "
                "'input_*' override settings are supplied."
            )
        },
    )
    input_prompt: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "Prefix to use as primary prompt for code lines when parsing input. Only used if "
                "'parsing_method' is 'declared'. If not set, 'prompt' is used for both input "
                "parsing and output formatting."
            )
        },
    )
    input_continuation: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "Prefix to use as secondary prompt for continued code lines when parsing input. "
                "Only used if 'parsing_method' is 'declared'. If not set, 'prompt' is used for "
                "both input parsing and output formatting."
            )
        },
    )
    input_comment: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "Prefix to use for results returned by expressions when parsing input. Only used "
                "if 'parsing_method' is 'declared'. If not set, 'prompt' is used for both input "
                "parsing and output formatting."
            )
        },
    )

    def __post_init__(self):
        # Validate venue
        if self.venue not in formatter_registry:
            raise InvalidVenueError(
                f"{self.venue} is not a valid value for parsing method."
                f"Valid values are: {list(formatter_registry.keys())}"
            )
        # Validate prompt and continuation prefixes
        if len(self.prompt) != len(self.continuation):
            raise PromptLengthMismatchError(
                f"Primary prompt ('{self.prompt}') and continuation prompt "
                f"('{self.continuation}') must be equal lengths."
            )
        # Validate parsing method
        try:
            ParsingMethod(self.parsing_method)
        except ValueError:
            raise InvalidParsingMethodError(
                f"{self.parsing_method} is not a valid value for parsing method."
                f"Valid values are: {[pm.value for pm in ParsingMethod]}"
            )

    @property
    def resolved_input_prompt(self):
        if self.input_prompt is not None:
            return self.input_prompt
        return self.prompt

    @property
    def resolved_input_continuation(self):
        if self.input_continuation is not None:
            return self.input_continuation
        return self.continuation

    @property
    def resolved_input_comment(self):
        if self.input_comment is not None:
            return self.input_comment
        return self.comment

    @classmethod
    def get_help(cls, field_name: str):
        return cls.__dataclass_fields__[field_name].metadata["help"]


# def format_args_google_style():
#     docs = []
#     for field in dataclasses.fields(ReprexConfig):
#         field_name = field.name
#         try:
#             field_type = field.type.__name__
#         except AttributeError:
#             field_type = str(field.type)
#         docs.extend(
#             textwrap.wrap(
#                 f"{field_name} ({field_type}): {CONFIG_DOCS[field_name]}",
#                 width=99,
#                 initial_indent=" " * 8,
#                 subsequent_indent=" " * 12,
#             )
#         )
#     return "\n".join(docs)[4:]


# if ReprexConfig.__doc__:
#     ReprexConfig.__doc__ = ReprexConfig.__doc__.replace("{{args}}", format_args_google_style())



class ReprexliteException(Exception):
    """Base class for reprexlite exceptions."""


class BlackNotFoundError(ModuleNotFoundError, ReprexliteException):
    """Raised when ipython cannot be found when using a black-dependent feature."""


class InputSyntaxError(SyntaxError, ReprexliteException):
    """Raised when encountering a syntax error when parsing input."""


class InvalidInputPrefixesError(ValueError, ReprexliteException):
    pass


class InvalidParsingMethodError(ValueError, ReprexliteException):
    pass


class InvalidVenueError(ValueError, ReprexliteException):
    pass


class IPythonNotFoundError(ModuleNotFoundError, ReprexliteException):
    """Raised when ipython cannot be found when using an IPython-dependent feature."""


class MissingDependencyError(ImportError, ReprexliteException):
    pass


class NoPrefixMatchError(ValueError, ReprexliteException):
    pass


class NotAFormatterError(TypeError, ReprexliteException):
    """Raised when registering a formatter that is not a subclass of the Formatter base class."""


class PromptLengthMismatchError(ReprexliteException):
    pass


class PygmentsNotFoundError(ModuleNotFoundError, ReprexliteException):
    """Raised when pygments cannot be found when using a pygments-dependent feature."""


class UnexpectedError(ReprexliteException):
    """Raised when an unexpected case happens."""

    def __init__(self, msg: str, *args: object):
        if not msg.endswith(" "):
            msg += " "
        msg += (
            "If you see this error from normal usage, please report at "
            "https://github.com/jayqi/reprexlite/issues"
        )
        super().__init__(msg, *args)



from reprexlite.exceptions import IPythonNotFoundError
from reprexlite.reprexes import Reprex, ReprexConfig, reprex
from reprexlite.version import __version__

try:
    from reprexlite.ipython import load_ipython_extension  # noqa: F401
except IPythonNotFoundError:
    pass

__all__ = [
    "Reprex",
    "ReprexConfig",
    "reprex",
]

__version__



from contextlib import contextmanager
import re
from typing import Optional

from typer.testing import CliRunner

import reprexlite.cli
from reprexlite.config import ReprexConfig
from reprexlite.exceptions import IPythonNotFoundError
from reprexlite.reprexes import Reprex
from reprexlite.version import __version__

try:
    from IPython import InteractiveShell
    from IPython.core.magic import Magics, line_cell_magic, magics_class
    from IPython.core.release import version as ipython_version
    from IPython.core.usage import default_banner_parts
    from IPython.terminal.interactiveshell import TerminalInteractiveShell
    from IPython.terminal.ipapp import TerminalIPythonApp
except ModuleNotFoundError as e:
    if e.name == "IPython":
        raise IPythonNotFoundError(*e.args, name="IPython")
    else:
        raise


runner = CliRunner()


@contextmanager
def patch_edit(input: str):
    """Patches typer.edit to return the input string instead of opening up the text editor. This
    is a trick to hook up the IPython cell magic's cell contents to the typer CLI app.
    """

    def return_input(*args, **kwargs) -> str:
        return input

    original = reprexlite.cli.typer.edit
    setattr(reprexlite.cli.typer, "edit", return_input)
    yield
    setattr(reprexlite.cli.typer, "edit", original)


@magics_class
class ReprexMagics(Magics):
    @line_cell_magic
    def reprex(self, line: str, cell: Optional[str] = None):
        """reprex IPython magic. Use line magic %reprex to print help. Use cell magic %%reprex to
        render a reprex."""
        # Line magic, print help
        if cell is None:
            help_text = runner.invoke(
                reprexlite.cli.app, ["--help"], env={"TERM": "dumb"}
            ).stdout.strip()
            help_text = re.sub(r"^Usage: main", r"Cell Magic Usage: %%reprex", help_text)
            print(f"reprexlite v{__version__} IPython Magic\n\n" + help_text)
            return
        # Cell magic, render reprex
        with patch_edit(cell):
            result = runner.invoke(reprexlite.cli.app, line.split())
            print(result.stdout, end="")


def load_ipython_extension(ipython: InteractiveShell):
    """Special function to register this module as an IPython extension.
    https://ipython.readthedocs.io/en/stable/config/extensions/#writing-extensions
    """

    ipython.register_magics(ReprexMagics)


ipython_banner_parts = [
    default_banner_parts[0],
    f"reprexlite {__version__} -- Interactive reprex editor via IPython {ipython_version}.",
]


class ReprexTerminalInteractiveShell(TerminalInteractiveShell):
    """Subclass of IPython's TerminalInteractiveShell that automatically executes all cells using
    reprexlite instead of normally."""

    banner1 = "".join(ipython_banner_parts)
    _reprex_config: Optional[ReprexConfig] = None

    def run_cell(self, raw_cell: str, *args, **kwargs):
        # "exit()" and Ctrl+D short-circuit this and don't need to be handled
        if raw_cell != "exit":
            try:
                r = Reprex.from_input(raw_cell, config=self.reprex_config)
                print(r.format(terminal=True), end="")
            except Exception as e:
                print("ERROR: reprexlite has encountered an error while evaluating your input.")
                print(e, end="")

            # Store history
            self.history_manager.store_inputs(self.execution_count, raw_cell, raw_cell)
            self.execution_count += 1

            return None
        else:
            return super().run_cell(raw_cell, *args, **kwargs)

    @property
    def reprex_config(self) -> ReprexConfig:
        if self._reprex_config is None:
            self._reprex_config = ReprexConfig()
        return self._reprex_config


class ReprexTerminalIPythonApp(TerminalIPythonApp):
    """Subclass of TerminalIPythonApp that launches ReprexTerminalInteractiveShell."""

    interactive_shell_class = ReprexTerminalInteractiveShell

    @classmethod
    def set_reprex_config(cls, config: ReprexConfig):
        """Set the reprex config bound on the interactive shell."""
        cls.interactive_shell_class._reprex_config = config
